//package src;

import java.util.*;

public class ActionRuleStrategy {
    private ArrayList<ArrayList<String>> data;
    private HashMap<ArrayList<String>, String> certainRules;
    private List<String> flexibleAttributes;
    private List<String> stableAttributes;
    private String decisionAttribute;
    private String decisionFrom;
    private String decisionTo;
    private ArrayList<ActionRule> actionRules;


    public ActionRuleStrategy(
        ArrayList<ArrayList<String>> data,
        HashMap<ArrayList<String>, String> certainRules, 
        List<String> flexibleAttributes,
        List<String> stableAttributes,
        String decisionAttribute,
        String decisionTo,
        String decisionFrom
    ) {
        this.decisionTo = decisionTo;
        this.decisionFrom = decisionFrom;
        this.flexibleAttributes = flexibleAttributes;
        this.stableAttributes = stableAttributes;
        this.decisionAttribute = decisionAttribute;
        this.certainRules = certainRules;
        this.data = data;
    }
    
    public void run(int min_support, float min_confidence) {
        this.actionRules = new ArrayList<ActionRule>();

        for(Map.Entry<ArrayList<String>,String> entry : this.certainRules.entrySet()) {
            HashMap<ArrayList<String>, String> certainRulesComplement 
                = new HashMap<ArrayList<String>, String>(this.certainRules);
            certainRulesComplement.remove(entry.getKey());
            for (ActionRule actionRule : this.findActionRules(entry, certainRulesComplement)) {
                this.actionRules.add(actionRule);
            }
        }

        ArrayList<ActionRule> tempActionRules = new ArrayList<ActionRule>();
        for(ActionRule actionRule : this.actionRules) {
            this.setActionRuleConfidence(actionRule);
            if (actionRule.getSupport() >= min_support 
                && actionRule.getConfidence() >= min_confidence
                && !tempActionRules.contains(actionRule)) {
                tempActionRules.add(actionRule);
            }
        }

        this.actionRules = tempActionRules;
    }

    public ArrayList<ActionRule> getActionRules() {
        return this.actionRules;
    }

    public ArrayList<ActionRule> findActionRules(
        Map.Entry<ArrayList<String>,String> entry, 
        HashMap<ArrayList<String>,String> certainRulesComplement
    ) {
        if(!entry.getValue().equals(this.decisionFrom) || !this.hasFlexibleAttributes(entry.getKey())) {
            return new ArrayList<ActionRule>();
        }

        ArrayList<ActionRule> actionRules = new ArrayList<ActionRule>();

        for(Map.Entry<ArrayList<String>,String> complementEntry : certainRulesComplement.entrySet()) {
            if(!complementEntry.getValue().equals(this.decisionTo) 
                || !this.hasFlexibleAttributes(complementEntry.getKey())) {
                continue;
            }

            ActionRule actionRule = findActionRule(entry.getKey(), complementEntry.getKey());
            if (actionRule != null) {
                actionRules.add(actionRule);
            }
        }

        return actionRules;
    }

    public boolean hasFlexibleAttributes(ArrayList<String> attributes) {
        for (int i = 0; i < attributes.size(); i ++) {
            for (int j = 0; j < this.flexibleAttributes.size(); j ++) {
                if (attributes.get(i).indexOf(this.flexibleAttributes.get(j)) > -1) {
                    return true;
                }
            }
        }

        return false;
    }

    public ArrayList<String> getFlexibleAttributes(ArrayList<String> attributes) {
        ArrayList<String> flexibleAttributeValues = new ArrayList<String>();

        for (int i = 0; i < attributes.size(); i ++) {
            for (int j = 0; j < this.flexibleAttributes.size(); j ++) {
                if (attributes.get(i).indexOf(this.flexibleAttributes.get(j)) > -1) {
                    flexibleAttributeValues.add(attributes.get(i));
                }
            }
        }
        return flexibleAttributeValues;
    }

    public ArrayList<String> getStableAttributes(ArrayList<String> attributes) {
        ArrayList<String> stableAttributeValues = new ArrayList<String>();

        for (int i = 0; i < attributes.size(); i ++) {
            for (int j = 0; j < this.stableAttributes.size(); j ++) {
                if (attributes.get(i).indexOf(this.stableAttributes.get(j)) > -1) {
                    stableAttributeValues.add(attributes.get(i));
                }
            }
        }

        return stableAttributeValues;
    }

    public ActionRule findActionRule(
        ArrayList<String> entryFrom, 
        ArrayList<String> entryTo
    ) {
        ArrayList<ArrayList<String>> flexibleTransition = this.findFlexibleTransition(entryFrom, entryTo);
        if (flexibleTransition.get(0).size() == 0) {
            return null;
        }

        ArrayList<String> stableTransition = this.findStableTransition(entryFrom, entryTo);
        return new ActionRule(
            flexibleTransition.get(0),
            flexibleTransition.get(1),
            stableTransition,
            this.decisionAttribute,
            this.decisionTo,
            this.decisionFrom
        );
    }

    public ArrayList<ArrayList<String>> findFlexibleTransition(
        ArrayList<String> entryFrom, 
        ArrayList<String> entryTo
    ) {
        ArrayList<String> transitionFrom = new ArrayList<String>();
        ArrayList<String> transitionTo = new ArrayList<String>();

        ArrayList<String> from = this.getFlexibleAttributes(entryFrom);
        ArrayList<String> to = this.getFlexibleAttributes(entryTo);


        for(int i = 0; i < from.size(); i++) {
            for(int j = 0; j < to.size(); j++) {
                // a == a && a_1 != a0
                if(this.doNamesMatch(from.get(i), to.get(j)) 
                    && !(from.get(i).equals(to.get(j)))) {
                    transitionFrom.add(from.get(i));
                    transitionTo.add(to.get(j));
                }
            }
        }

        return new ArrayList<ArrayList<String>>() { 
            { 
                add(transitionFrom); 
                add(transitionTo);
            } 
        };
    }

    public ArrayList<String> findStableTransition(
        ArrayList<String> entryFrom, 
        ArrayList<String> entryTo
    ) {
        ArrayList<String> stableAttributes = new ArrayList<String>();

        ArrayList<String> from = this.getStableAttributes(entryFrom);
        ArrayList<String> to = this.getStableAttributes(entryTo);

        for(int i = 0; i < from.size(); i++) {
            for(int j = 0; j < to.size(); j++) {
                // a == a && a0 == a0
                if(this.doNamesMatch(from.get(i), to.get(j)) 
                    && (from.get(i).equals(to.get(j)))
                    && stableAttributes.indexOf(from.get(i)) == -1) {
                    stableAttributes.add(from.get(i));
                }
            }
        }

        return stableAttributes;
    }

    private boolean doNamesMatch(String from, String to) {
        return from.split(IOHandler.ATTRIBUTE_DELIMITER)[0].equals(
            to.split(IOHandler.ATTRIBUTE_DELIMITER)[0]);
    }

    public void setActionRuleConfidence(ActionRule actionRule) {
        int y1_z1Count = 0;
        int y2_z2Count = 0;
        int y1Count = 0;
        int y2Count = 0;
        ArrayList<String> y1_z1 = actionRule.getY1Z1();
        ArrayList<String> y2_z2 = actionRule.getY2Z2();
        ArrayList<String> y1 = actionRule.getY1();
        ArrayList<String> y2 = actionRule.getY2();
        for(ArrayList<String> dataEntry : data) {
            if(dataEntry.containsAll(y1)) {
                y1Count++;
            }
            if(dataEntry.containsAll(y2)) {
                y2Count++;
            }
            if(dataEntry.containsAll(y1_z1)) {
                y1_z1Count++;
            }
            if(dataEntry.containsAll(y2_z2)) {
                y2_z2Count++;
            }
        }
        if(y1Count == 0 || y2Count ==0) {
            return;
        }
        actionRule.setSupport(y1_z1Count);
        actionRule.setConfidence(((y1_z1Count * 1.0f)/(y1Count * 1.0f)) * ((y2_z2Count * 1.0f)/(y2Count * 1.0f)));
    }
    
    public static void test() {
        HashMap<ArrayList<String>, String> certainRules = new HashMap<ArrayList<String>, String>();
        certainRules.put(new ArrayList<String>() { { add("a_1"); add("b_1"); add("c_1"); } }, "d_1");
        certainRules.put(new ArrayList<String>() { { add("a_2"); add("b_1"); add("c_1"); } }, "d_2");
        certainRules.put(new ArrayList<String>() { { add("a_1"); add("b_2"); add("c_1"); } }, "d_2");
        certainRules.put(new ArrayList<String>() { { add("a_2"); add("b_2"); add("c_2"); } }, "d_2");
        certainRules.put(new ArrayList<String>() { { add("a_1"); add("b_1"); add("c_1"); } }, "d_1");
        List<String> flexibleAttributes = Arrays.asList("a", "b");
        List<String> stableAttributes = Arrays.asList("c");
        String decisionAttribute = "d";
        String decisionFrom = "d_1";
        String decisionTo = "d_2";

        ArrayList<ArrayList<String>> data = new ArrayList<ArrayList<String>>() { 
            { 
                add(new ArrayList<String>() { { add("a_1"); add("b_1"); add("c_2"); add("d_1"); } });
                add(new ArrayList<String>() { { add("a_2"); add("b_1"); add("c_1"); add("d_2"); } });
                add(new ArrayList<String>() { { add("a_1"); add("b_1"); add("c_2"); add("d_2"); } });
                add(new ArrayList<String>() { { add("a_2"); add("b_2"); add("c_1"); add("d_2"); } });
                add(new ArrayList<String>() { { add("a_1"); add("b_2"); add("c_2"); add("d_1"); } });
                add(new ArrayList<String>() { { add("a_2"); add("b_2"); add("c_1"); add("d_1"); } });
            } 
        };
        ActionRuleStrategy strategy = new ActionRuleStrategy(
            data,
            certainRules,
            flexibleAttributes,
            stableAttributes,
            decisionAttribute,
            decisionFrom,
            decisionTo);

        assert(strategy.hasFlexibleAttributes(new ArrayList<String>() { { add("a_1"); add("b_1"); add("c_1"); } }) == true);
        assert(strategy.hasFlexibleAttributes(new ArrayList<String>() { { add("c_1"); } }) == false);
        ArrayList<String> tempFlexAttr = strategy.getFlexibleAttributes(new ArrayList<String>() { { add("a_1"); add("b_1"); add("c_1"); } });
        assert(tempFlexAttr.indexOf("a_1") > -1);
        assert(tempFlexAttr.indexOf("b_1") > -1);
        assert(tempFlexAttr.indexOf("c_1") == -1);

        ArrayList<String> tempStableAttr = strategy.getStableAttributes(new ArrayList<String>() { { add("a_1"); add("b_1"); add("c_1"); } });
        assert(tempStableAttr.indexOf("a_1") == -1);
        assert(tempStableAttr.indexOf("b_1") == -1);
        assert(tempStableAttr.indexOf("c_1") > -1);

        ArrayList<ArrayList<String>> flexibleTransition = strategy.findFlexibleTransition(
            new ArrayList<String>() { { add("a_1"); add("b_1"); add("c_1"); } }, // from
            new ArrayList<String>() { { add("a_2"); add("b_2"); add("c_1"); } }  // to
        );
        assert(flexibleTransition.get(0).indexOf("a_1") > -1);
        assert(flexibleTransition.get(0).indexOf("b_1") > -1);
        assert(flexibleTransition.get(0).indexOf("c_1") == -1);
        assert(flexibleTransition.get(1).indexOf("a_2") > -1);
        assert(flexibleTransition.get(1).indexOf("b_2") > -1);
        assert(flexibleTransition.get(1).indexOf("c_1") == -1);

        ArrayList<String> stableTransition = strategy.findStableTransition(
            new ArrayList<String>() { { add("a_1"); add("b_1"); add("c_1"); } }, // from
            new ArrayList<String>() { { add("a_2"); add("b_2"); add("c_1"); } }  // to
        );
        assert(flexibleTransition.indexOf("a_1") == -1);
        assert(flexibleTransition.indexOf("a_2") == -1);
        assert(flexibleTransition.indexOf("b_1") == -1);
        assert(flexibleTransition.indexOf("b_2") == -1);
        assert(flexibleTransition.indexOf("c_1") > -1);
        ActionRule actionRule = strategy.findActionRule(
            new ArrayList<String>() { { add("a_1"); add("b_1"); add("c_1"); } }, // from
            new ArrayList<String>() { { add("a_2"); add("b_2"); add("c_1"); } }  // to
        );
        assert(actionRule.toString().equals("[(a, a_1 → a_1) ^ (b, b_1 → b_1) ^ c_1] → (d, d_1 → d_2) = (0, 0)"));
        strategy.run(0, 0.0f);
        ArrayList<ActionRule> actionRules = strategy.getActionRules();
        for (int i = 0; i < actionRules.size(); i ++) {
            System.out.println(actionRules.get(i).toString());
        }

        System.out.println("The ActionRuleStrategy test cases passed");
    }
}